/*
 * Copyright (C) 2017-2020 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "strings.hpp"

namespace {
bool u8_1b(const uint8_t *bp)
{
    return bp[0] == 0x09 || bp[0] == 0x0a || bp[0] == 0x0d || (bp[0] >= 0x20 && bp[0] <= 0x7e);
}

bool u8_2b(const uint8_t *bp)
{
     return bp[0] >= 0xc2 && bp[0] <= 0xdf && bp[1] >= 0x80 && bp[1] <= 0xbf;
}

bool u8_3b(const uint8_t *bp)
{
    if(bp[0] == 0xe0 && bp[1] >= 0xa0 && bp[1] <= 0xbf && bp[2] >= 0x80 && bp[2] <= 0xbf)
        return true;

    if(bp[0] >= 0xe1 && bp[0] <= 0xef && bp[0] != 0xed && bp[1] >=0x80 && bp[1] <= 0xbf && bp[2] >= 0x80 && bp[2] <= 0xbf)
        return true;

    if(bp[0] == 0xed && bp[1] >= 0x80 && bp[1] <= 0x9f && bp[2] >= 0x80 && bp[2] < 0xbf)
        return true;

    return false;
}

bool u8_4b(const uint8_t *bp)
{
    if(bp[0] == 0xf0 && bp[1] >= 0x90 && bp[1] <= 0xbf && bp[2] >= 0x80 && bp[2] <= 0xbf && bp[3] >= 0x80 && bp[3] <= 0xbf)
        return true;

    if(bp[0] >= 0xf1 && bp[0] <= 0xf3 && bp[1] >= 0x80 && bp[1] <= 0xbf && bp[2] >= 0x80 && bp[2] <= 0xbf && bp[3] >= 0x80 && bp[3] <= 0xbf)
        return true;

    if(bp[0] == 0xf4 && bp[1] >= 0x80 && bp[1] <= 0x8f && bp[2] >= 0x80 && bp[2] <= 0xbf && bp[3] >= 0x80 && bp[3] <= 0xbf)
        return true;

    return false;
}
}

size_t u8verify(const char *cp, size_t max)
{
    if(!cp || !*cp || !max)
        return 0;

    size_t size = 0, codepoints = 0;
    const uint8_t *bp = reinterpret_cast<const uint8_t *>(cp);
    while(*bp && codepoints < max) {
        if(u8_1b(bp)) {
            ++bp;
            ++size;
            ++codepoints;
            continue;
        }

        if(u8_2b(bp)) {
            size += 2;
            bp +=2;
            ++codepoints;
            continue;
        }

        if(u8_3b(bp)) {
            size += 3;
            bp += 3;
            ++codepoints;
            continue;
        }


        if(u8_4b(bp)) {
            size += 4;
            bp += 4;
            ++codepoints;
            continue;
        }

        return 0;
    }

    if(codepoints > max || *bp)
        return 0;

    return size;
}

std::vector<std::string> split(const std::string& str, const std::string& delim)
{
    std::vector<std::string> result;
    std::size_t current, prev = 0;
    current = str.find_first_of(delim);
    while(current != std::string::npos) {
        result.push_back(str.substr(prev, current - prev));
        prev = current + 1;
        current = str.find_first_of(delim, prev);
    }
    result.push_back(str.substr(prev, current - prev));
    return result;
}

const char *get_env(const char *id, size_t max)
{
    const char *cp = ::getenv(id);  // FlawFinder: verified input
    if(!cp)
        return nullptr;

    if(!u8verify(cp, max))
        return nullptr;

    return cp;
}
